class UE_randomisation_EventHandler : EventHandler
{
	Actor player;

	override void WorldLoaded(WorldEvent e)
	{
		super.WorldLoaded(e);

		self.player = players[consoleplayer].mo;
		Console.Printf("%s", "Universal Entropy 3.666b loaded");
	}

	override void WorldThingSpawned(WorldEvent e)
	{
		super.WorldThingSpawned(e);

		Actor a = e.Thing;
		if(!a) { Return; }
		if ((!ue_friend && (a.bFriendly || a.IsFriend(self.player) || a.isTeammate(self.player) || !a.isHostile(self.player))) || (a is "PlayerPawn")) { return; }

		double size_mean = 1. + (ue_bias - 2.) / 8.;
		double size1 = frandom(size_mean - ue_rr, size_mean + ue_rr);
		double size2 = frandom(size_mean - ue_rr, size_mean + ue_rr);

		// ratio preservation
		if (ue_rp > 0 && ue_rp < 1.01) // ratio preservation [0..1]
		{
			size2 += (size1 - size2) * ue_rp;
		}

		double size = size1 * size2;
		double szsq = size1 * size1;
		double szrt = sqrt(size);

		// Monster randomisation
		if (a.bIsMonster)
		{
			// Resizing
			if ((ue_resize > 0) && (ue_resize >= random(1, 5)))
			{
				bool resized = a.A_SetSize(a.Radius * size1, a.Height * size2, true);
				if (resized)
				{
					a.Health = round(size * a.Health);
					a.Mass = round(size * a.Mass);
					a.A_SetScale(a.Scale.X * size1, a.Scale.Y * size2);

					a.PainChance = round(a.PainChance / szrt);
					a.MeleeRange *= szrt;
					a.PushFactor /= szrt;

					a.SetDamage(round(szrt * a.Damage));
					a.DamageMultiply *= szrt;

					a.A_GiveInventory('Shifter');
					let shifter = Shifter(a.FindInventory('Shifter'));
					if (shifter != null)
					{
						shifter.size = szrt;
						shifter.shift_speed = 1;
					}
				}
			}

			// Recolouring
			if ((ue_recolour > 0) && (ue_recolour >= random(1, 5)))
			{
				a.Translation = ColorTranslationWrapper.GenerateRandomColorTranslation();
			}
		}
		else
		{
			// Decoration randomisation
			if (a.bSolid)
			{
				if ((ue_decor_resize > 0) && (ue_decor_resize >= random(1, 5)))
				{
					bool resized = a.A_SetSize(a.Radius * size1, a.Height * size2, true);
					if (resized) { a.A_SetScale(a.Scale.X * size1, a.Scale.Y * size2); }

					a.A_GiveInventory('Shifter');
					let shifter = Shifter(a.FindInventory('Shifter'));
					if (shifter != null)
					{
						shifter.size = szrt;
						shifter.shift_speed = 0;
					}
				}

				if ((ue_decor_recolour > 0) && (ue_decor_recolour >= random(1, 5)))
				{
					a.Translation = ColorTranslationWrapper.GenerateRandomColorTranslation();
				}
			}
		}

		// Ammo pack randomisation
		if ((a is 'Ammo' && Inventory(a).Amount > 4) && (ue_ammo > 0) && (ue_ammo >= random(1, 5)))
		{
			bool resized = a.A_SetSize(a.Radius * size1, a.Height * size1, true);
			if (resized)
			{
				Inventory(a).Amount = round(szsq * Inventory(a).Amount);
				a.A_SetScale(a.Scale.X * size1, a.Scale.Y * size1);
			}
		}

		// Health kit randomisation
		if ((a is 'Health' && Inventory(a).Amount > 4 && Inventory(a).Amount < 100) && (ue_healthkit > 0) && (ue_healthkit >= random(1, 5)))
		{
			bool resized = a.A_SetSize(a.Radius * size1, a.Height * size1, true);
			if (resized)
			{
				Inventory(a).Amount = round(szsq * Inventory(a).Amount);
				a.A_SetScale(a.Scale.X * size1, a.Scale.Y * size1);
			}
		}

		// Armour randomisation
		if ((a is 'BasicArmorPickup') && (ue_armour > 0) && (ue_armour >= random(1, 5)))
		{
			bool resized = a.A_SetSize(a.Radius * size1, a.Height * size1, true);
			if (resized)
			{
				BasicArmorPickup(a).SaveAmount = round(szsq * BasicArmorPickup(a).SaveAmount);
				a.A_SetScale(a.Scale.X * size1, a.Scale.Y * size1);
			}
		}

		// Missile randomisation
		if ((a.bMissile) && (a.target != null) && !(a.target is "PlayerPawn"))
		{
			a.Vel.xy = Actor.RotateVector(a.Vel.xy, frandom(-11 * ue_rr, 11 * ue_rr));

			let shifter = Shifter(a.target.FindInventory('Shifter'));
			if (shifter != null)
			{
				a.Vel.xy /= shifter.size;
				a.Mass = round(shifter.size * a.Mass);
				a.A_SetScale(a.Scale.X * shifter.size, a.Scale.Y * shifter.size);
			}
		}
	} // end WorldThingSpawned()
} // end UE_randomisation_EventHandler

class Shifter : Inventory
{
	double size;
	state ownerstate;
	bool shift_speed;

	override void Tick()
	{
		Super.Tick();

		if (Owner != null)
		{
			Owner.A_SoundPitch(CHAN_VOICE, 1 / size);
			Owner.A_SoundPitch(CHAN_WEAPON, 1 / size);
			if (shift_speed && Owner.Tics >= 1 && ownerstate != Owner.curState)
			{
				Owner.A_SetTics(round(Owner.tics * size));
				ownerstate = Owner.curState;
			}
		}
	}
}
